#include "BaseAttacker.h"
#include "FindModule.h"
#include "utilfunctions.h"
#include "MovingMobilityBase.h"

Define_Module(BaseAttacker);

BaseAttacker::BaseAttacker()
    : BaseModule()
    , debug(false)
    , numNodes(0)
    , numAttackerNodes(0)
    , attackerPhaseStart(0)
    , attackerPhaseEnd(0)
    , attackerCompRate(0)
    , attackerRange(0)
    , numInfectedNodes(0)
    , capturedKeys()
    , capturedNonces()
    , infectedNodes()
{}

void BaseAttacker::initialize(int stage) {
    BaseModule::initialize(stage);

    if (stage == 0) {
        if (hasPar("debug")) debug = par("debug").boolValue();
        if (hasPar("numNodes")) numNodes = par("numNodes");
        if (hasPar("numAttackerNodes")) numAttackerNodes = par("numAttackerNodes");
        if (hasPar("attackerPhaseStart")) attackerPhaseStart = par("attackerPhaseStart");
        if (hasPar("attackerPhaseEnd")) attackerPhaseEnd = par("attackerPhaseEnd");
        if (hasPar("attackerCompRate")) attackerCompRate = par("attackerCompRate");
        if (hasPar("attackerRange")) attackerRange = par("attackerRange");
        if (hasPar("numInfectedNodes")) numInfectedNodes = par("numInfectedNodes");

    } else if (stage == 1) {
        // In second initialization stage, schedule the start and end of attacker operation
        cMessage * attackerPhaseStartTimer = new cMessage("attackerPhaseStartTimer", ATTACKER_START);
        attackerPhaseStartTimer->setSchedulingPriority(1);
        scheduleAt(attackerPhaseStart, attackerPhaseStartTimer);

        cMessage * attackerPhaseEndTimer = new cMessage("attackerPhaseEndTimer", ATTACKER_END);
        scheduleAt(attackerPhaseEnd, attackerPhaseEndTimer);
    }
}

void BaseAttacker::handleMessage(cMessage *msg) {
    if (msg->getKind() == ATTACKER_START) {
        performCompromitation();
        checkLinksStatistics();
    } else if (msg->getKind() == ATTACKER_END) {
        checkLinksStatistics();
    }
    delete msg;
}

void BaseAttacker::attackerKeyUpdateCheck(Key oldKey, Nonce nonce, Key newKey) {
    if (capturedKeys.find(oldKey) != capturedKeys.end()) {
        if (capturedNonces.find(nonce) != capturedNonces.end()) {
            if ((capturedKeys.insert(newKey)).second) {
                emit(attackerKeyStoredID, 1);
            }
        }
    }

    emit(attackerKeyUpdateID, 1);
}

void BaseAttacker::attackerNonceSentCheck(Address nodeAddress, Key key, Nonce nonce, Coord nodePosition) {
    if (isMalwareInfected(nodeAddress)) {
        if ((capturedNonces.insert(nonce)).second) {
            emit(attackerNonceStoredID, 1);
        }
    } else if (isInAttackerRange(nodePosition)) {
        if (capturedKeys.find(key) != capturedKeys.end()) {
            if ((capturedNonces.insert(nonce)).second) {
                emit(attackerNonceStoredID, 1);
            }
        }
    }
}

void BaseAttacker::attackerNonceReceivedCheck(Address nodeAddress, Nonce nonce) {
    if (isMalwareInfected(nodeAddress)) {
        if ((capturedNonces.insert(nonce)).second) {
            emit(attackerNonceStoredID, 1);
        }
    }
}

bool BaseAttacker::isInAttackerRange(Coord nodePosition) {
    // Zero means that attacker has unlimited range
    if (attackerRange == 0) {
        return true;
    }

    // Check the node position against all attacker nodes
    for (int i = 0; i < numAttackerNodes; i++) {
        cModule* attackerNode = FindModule<>::findNetwork(this)->getSubmodule("attackerNode", i);
        MovingMobilityBase* mobility = check_and_cast<MovingMobilityBase *>(attackerNode->getSubmodule("mobility"));
        Coord attackerNodePosition = mobility->getCurrentPosition();

        if (attackerNodePosition.distance(nodePosition) <= attackerRange) {
            return true;
        }
    }

    return false;
}

bool BaseAttacker::isMalwareInfected(Address address) {
    if (infectedNodes.find(address) == infectedNodes.end()) {
        return 0;
    } else {
        return 1;
    }
}

void BaseAttacker::checkLinksStatistics() {
    int linksTotal = 0;
    int linksWithKey = 0;
    int compromisedLinks = 0;

    for (int i = 0; i < numNodes; i++) {
        cModule * nodeA = FindModule<>::findNetwork(this)->getSubmodule("node", i);
        Data * dataA = check_and_cast<Data *>(nodeA->getSubmodule("kms")->getSubmodule("data"));

        for (int j = i+1; j < numNodes; j++) {
            cModule * nodeB = FindModule<>::findNetwork(this)->getSubmodule("node", j);
            Data * dataB = check_and_cast<Data *>(nodeB->getSubmodule("kms")->getSubmodule("data"));

            if (areNodesNeighbours(nodeA, nodeB)) {

                // Nodes A and B are neighbors
                linksTotal++;

                if (existsSharedKeyBetween(nodeA->getIndex(), dataA, nodeB->getIndex(), dataB)) {
                    // Nodes A and B shares mutual key
                    linksWithKey++;

                    Key key = dataA->getKey(nodeB->getIndex());

                    if (capturedKeys.find(key) != capturedKeys.end()) {
                        // Shared key is compromised
                        compromisedLinks++;
                    }
                }
            }
        }
    }

    emit(totalLinkCountID, linksTotal);
    emit(linksWithKeyCountID, linksWithKey);
    emit(comprLinkCountID, compromisedLinks);
    emit(comprLinkRateID, (double)compromisedLinks / linksWithKey * 100);
}

void BaseAttacker::performCompromitation() {

}

BaseAttacker::~BaseAttacker() {

}
