#ifndef BASE_ATTACKER_H
#define BASE_ATTACKER_H

#include "kmsdefs.h"
#include "BaseModule.h"
#include "Coord.h"

/**
 * @brief Base class for implementation of attacker
 *
 * The common functions for all attackers are implemented. This includes
 * start and stop attacker period, functions to verify whether the nonce
 * or the key were compromised. Also several help functions are
 * implemented including computation of statistics and more.
 *
 * @author Radim Ostadal
 **/
class MIXIM_API BaseAttacker : public BaseModule {

    protected:
        /** @brief Debug switch */
        bool debug;

        /** @brief Number of legitimate nodes in the network */
        int numNodes;

        /** @brief Number of malicious nodes in the network */
        int numAttackerNodes;

        /** @brief Attacker starts her activity at that time */
        int attackerPhaseStart;

        /** @brief Attacker ends her activity at that time */
        int attackerPhaseEnd;

        /** @brief Rate of links that will be compromised [0, 1] */
        double attackerCompRate;

        /** @brief Range in meters, attacker eavesdrop all communication in that range */
        double attackerRange;

        /** @brief Number of nodes that are infected by a malware */
        int numInfectedNodes;

        /** @brief Stores all keys compromised/captured by the attacker */
        KeySet capturedKeys;

        /** @brief Stores all nonces compromised/captured by the attacker */
        NonceSet capturedNonces;

        /** @brief Stores all nodes infected by a malware */
        NodesSet infectedNodes;

        // All signals for collection of statistics
        simsignal_t totalLinkCountID = registerSignal("totalLinkCount");
        simsignal_t linksWithKeyCountID = registerSignal("linksWithKeyCount");
        simsignal_t comprLinkCountID = registerSignal("comprLinkCount");
        simsignal_t comprLinkRateID = registerSignal("comprLinkRate");
        simsignal_t comprNodeID = registerSignal("comprNode");
        simsignal_t infectedNodeID = registerSignal("infectedNode");
        simsignal_t attackerKeyUpdateID = registerSignal("attackerKeyUpdate");
        simsignal_t attackerKeyStoredID = registerSignal("attackerKeyStored");
        simsignal_t attackerNonceStoredID = registerSignal("attackerNonceStored");

        /** @breif Message types used for timers */
        enum selfMessages {
            ATTACKER_START,
            ATTACKER_END,
        };

    private:
        /** @brief Copy constructor is not allowed */
        BaseAttacker (const BaseAttacker&);

        /** @brief Assignment operator is not allowed */
        BaseAttacker& operator=(const BaseAttacker&);

    public:
        BaseAttacker();
        virtual ~BaseAttacker();

        /** @brief Initialization of variables, activity planned using timers */
        virtual void initialize(int stage);

        /** @brief Handle the self messages - start and end attacker activity */
        virtual void handleMessage(cMessage *msg);

        /**
         * @brief The function computes statistics for the attacker
         *
         * Statistics includes number of links, number of links with key,
         * number of compromised links and its respective rate.
         **/
        void checkLinksStatistics();

        /** @brief Returns the true if the node is in range of at least one attacker node, false otherwise */
        bool isInAttackerRange(Coord nodePosition);

        /** @brief Returns the true if the node is infected by malware, false otherwise */
        bool isMalwareInfected(Address address);

        /**
         * @brief This function is used before the node update his link key
         *
         * Attacker has to have access to both original link key before the
         * update and also the nonce that is used to update the key. If so,
         * the attacker is able to compromise and store the updated key.
         *
         * @param oldKey The link key before the update
         * @param nonce The nonce that is used to update the key
         * @param newKey The link key after the update
         **/
        void attackerKeyUpdateCheck(Key oldKey, Key nonce, Key newKey);

        /**
         * @brief This function is used before the nonce is sent to its neighbour
         *
         * To successfully compromise the nonce being sent, the attacker have to fulfil
         * following criteria: node is infected by malware, or the originating node
         * (nodeAddress) has to be both in attacker eavesdropping range and the attacker
         * has to have access to used link key.
         *
         * @param nodeAddress Address of originating node, necessary to identify whether the node is infected by malware
         * @param key Key used for next hop neighbour, the key is used to encrypt the message
         * @param nonce The nonce being sent in the message (will be stored by attacker in case of successful compromise)
         * @param nodePosition Coordinates of originating node, necessary to identify whether the node is in range of attacker
         **/
        void attackerNonceSentCheck(Address nodeAddress, Key key, Key nonce, Coord nodePosition);

        /**
         * @brief This function is used after the nonce is received from its neighbour
         *
         * If the node is infected by malware, the nonce is stored by the attacker.
         *
         * @param nodeAddress Address of originating node, necessary to identify whether the node is infected by malware
         * @param nonce The nonce being sent in the message (will be stored by attacker in case of successful compromise)
         **/
        void attackerNonceReceivedCheck(Address nodeAddress, Key nonce);

        /** @brief Return all keys compromised/captured by the attacker */
        KeySet getCapturedKeys() {return capturedKeys;};

        /** @brief Return all nonces compromised/captured by the attacker */
        KeySet getCapturedNonces() {return capturedNonces;};

        /** @brief Virtual function that has to be implemented in actual class performing initial compromise */
        virtual void performCompromitation();
};

#endif
