#include "NodeCompAttacker.h"
#include "FindModule.h"
#include "utilfunctions.h"
#include "BaseWorldUtility.h"
#include "StaticMobility.h"

Define_Module(NodeCompAttacker);

NodeCompAttacker::NodeCompAttacker()
    : BaseAttacker()
    , attackerCompPattern(0)
    , compromisedNodes()
{}

void NodeCompAttacker::initialize(int stage) {
    BaseAttacker::initialize(stage);
    if (stage == 0) {
        if (hasPar("attackerCompPattern")) {
            if (!strcmp(par("attackerCompPattern"), "RANDOM")) {attackerCompPattern = RANDOM_PATTERN;}
            if (!strcmp(par("attackerCompPattern"), "OUTERMOST")) {attackerCompPattern = OUTERMOST_PATTERN;}
            if (!strcmp(par("attackerCompPattern"), "DIRECT_CENTER")) {attackerCompPattern = DIRECT_CENTER_PATTERN;}
            if (!strcmp(par("attackerCompPattern"), "CENTER_DROP")) {attackerCompPattern = CENTER_DROP_PATTERN;}
        }
    } else if (stage == 1) {

    }
}

void NodeCompAttacker::performCompromitation() {

    int linksWithKey = 0;

    // Count number of keys in the network
    for (int i = 0; i < numNodes; i++) {
        cModule * nodeA = FindModule<>::findNetwork(this)->getSubmodule("node", i);
        Data * dataA = check_and_cast<Data *>(nodeA->getSubmodule("kms")->getSubmodule("data"));

        for (int j = i+1; j < numNodes; j++) {
            cModule * nodeB = FindModule<>::findNetwork(this)->getSubmodule("node", j);
            Data * dataB = check_and_cast<Data *>(nodeB->getSubmodule("kms")->getSubmodule("data"));

            if (areNodesNeighbours(nodeA, nodeB)) {
                if (existsSharedKeyBetween(nodeA->getIndex(), dataA, nodeB->getIndex(), dataB)) {
                    // Nodes A and B shares mutual key
                    linksWithKey++;
                }
            }
        }
    }

    // Count the relevant number of keys that should be compromised
    int keysToBeCaptured = attackerCompRate * linksWithKey;
    EV << "Number of keys to be compromised:" << keysToBeCaptured << endl;

    // Count the coordinates of the Center
    BaseWorldUtility * world = check_and_cast<BaseWorldUtility *>(FindModule<>::findNetwork(this)->getSubmodule("world"));
    Coord * centre = new Coord(world->getPgs()->x/2, world->getPgs()->y/2, world->getPgs()->z/2);

    // Execute particular compromising protocol according to attackerCompPattern parameter
    if (attackerCompPattern == RANDOM_PATTERN) {
        compromisedNodes = randomCompromitation(keysToBeCaptured);
    } else if (attackerCompPattern == OUTERMOST_PATTERN) {
        compromisedNodes = outermostCompromitation(keysToBeCaptured, centre);
    } else if (attackerCompPattern == DIRECT_CENTER_PATTERN) {
        compromisedNodes = directCenterCompromitation(keysToBeCaptured, centre);
    } else if (attackerCompPattern == CENTER_DROP_PATTERN) {
        compromisedNodes = centerDropCompromitation(keysToBeCaptured, centre);
    }

    // Load a malware application to the random set of compromised nodes
    // The maximum number of infected nodes is the number of nodes compromised
    if (numInfectedNodes > compromisedNodes.size()) {
        numInfectedNodes = compromisedNodes.size();
    }

    double rateInfectedNodes = double(numInfectedNodes) / double(compromisedNodes.size());

    for (NodesSet::iterator it = compromisedNodes.begin(); infectedNodes.size() < numInfectedNodes; it++) {

        if (it == compromisedNodes.end()) {
            it = compromisedNodes.begin();
        }

        if (uniform(0, 1) < rateInfectedNodes) {

            // Perform compromitation
            if ((infectedNodes.insert(*it)).second) {
                emit(infectedNodeID, 1);
            }
        }
    }
}

NodesSet NodeCompAttacker::randomCompromitation(int keysToBeCaptured) {

    NodesSet selectedNodes;

    for (int i = 0; capturedKeys.size() < keysToBeCaptured; i++) {

        if (i == numNodes) {
            i = 0;
        }

        if (uniform(0, 1, 5) < attackerCompRate) {
            cModule * node = FindModule<>::findNetwork(this)->getSubmodule("node", i);
            Data * data = check_and_cast<Data *>(node->getSubmodule("kms")->getSubmodule("data"));
            Keys keys = data->getKeys();

            if ((selectedNodes.insert(i)).second) {
                emit(comprNodeID, 1);
            }

            for (Keys::iterator it = keys.begin(); it != keys.end(); it++) {
                if ((capturedKeys.insert(it->second)).second) {
                    emit(attackerKeyStoredID, 1);
                }
            }
        }
    }

    return selectedNodes;
}

NodesSet NodeCompAttacker::outermostCompromitation(int keysToBeCaptured, Coord * centre) {

    NodesSet selectedNodes;
    std::multimap<double, int> mapOfDistances;

    for (int i = 0; i < numNodes; i++) {
        cModule * node = FindModule<>::findNetwork(this)->getSubmodule("node", i);
        StationaryMobility * mobility = check_and_cast<StationaryMobility *>(node->getSubmodule("mobility"));
        Coord position = mobility->getCurrentPosition();

        mapOfDistances.insert(std::pair<double,int>(position.distance(*centre), i));
    }

    for (std::multimap<double, int>::reverse_iterator it = mapOfDistances.rbegin(); capturedKeys.size() < keysToBeCaptured && it != mapOfDistances.rend(); it++) {

        cModule * node = FindModule<>::findNetwork(this)->getSubmodule("node", it->second);
        Data * data = check_and_cast<Data *>(node->getSubmodule("kms")->getSubmodule("data"));
        Keys keys = data->getKeys();

        if ((selectedNodes.insert(it->second)).second) {
            emit(comprNodeID, 1);
        }

        for (Keys::iterator it = keys.begin(); it != keys.end(); it++) {
            if ((capturedKeys.insert(it->second)).second) {
                emit(attackerKeyStoredID, 1);
            }
        }

        // For debug purpose only
        StationaryMobility * mobility = check_and_cast<StationaryMobility *>(node->getSubmodule("mobility"));
        Coord position = mobility->getCurrentPosition();
        EV << "\t Compromising node with coordinates: [" << position.x << ", " << position.y << "], distance: " << position.distance(*centre) << endl;
    }

    delete centre;
    return selectedNodes;
}

NodesSet NodeCompAttacker::directCenterCompromitation(int keysToBeCaptured, Coord * centre) {

    NodesSet selectedNodes;
    std::multimap<double, int> mapOfDistances;

    // Select random starting point on the boundary
    BaseWorldUtility * world = check_and_cast<BaseWorldUtility *>(FindModule<>::findNetwork(this)->getSubmodule("world"));
    double relativStart = uniform(0, (world->getPgs()->x*2) + (world->getPgs()->y*2));
    Coord * start;

    if (relativStart <= world->getPgs()->x) {
        // Point in on bottom side
        start = new Coord (relativStart, 0, 0);
    } else if (relativStart <= world->getPgs()->x + world->getPgs()->y) {
        // Point in on right side
        start = new Coord (world->getPgs()->x, relativStart - world->getPgs()->x, 0);
    } else if (relativStart <= (world->getPgs()->x*2) + world->getPgs()->y) {
        // Point in on top side
        start = new Coord (relativStart - world->getPgs()->x - world->getPgs()->y, world->getPgs()->y, 0);
    } else {
        // Point in on left side
        start = new Coord (0, relativStart - world->getPgs()->y - (world->getPgs()->x*2), 0);
    }

    for (int i = 0; i < numNodes; i++) {
        cModule * node = FindModule<>::findNetwork(this)->getSubmodule("node", i);
        StationaryMobility * mobility = check_and_cast<StationaryMobility *>(node->getSubmodule("mobility"));
        Coord position = mobility->getCurrentPosition();

        // Compute distance
        double tempDistance = (*start).distance(*centre);
        double t = ((position.x - (*start).x) * ((*centre).x - (*start).x) + (position.y - (*start).y) * ((*centre).y - (*start).y)) / (tempDistance*tempDistance);
        t = fmax(0, fmin(1, t));
        Coord projection ((*start).x + t * ((*centre).x - (*start).x), (*start).y + t * ((*centre).y - (*start).y), 0);
        double distance = position.distance(projection);

        mapOfDistances.insert(std::pair<double, int>(distance, i));
    }

    for (std::multimap<double, int>::iterator it = mapOfDistances.begin(); capturedKeys.size() < keysToBeCaptured && it != mapOfDistances.end(); it++) {
        cModule * node = FindModule<>::findNetwork(this)->getSubmodule("node", it->second);
        Data * data = check_and_cast<Data *>(node->getSubmodule("kms")->getSubmodule("data"));
        Keys keys = data->getKeys();

        if ((selectedNodes.insert(it->second)).second) {
            emit(comprNodeID, 1);
        }

        for (Keys::iterator it = keys.begin(); it != keys.end(); it++) {
            if ((capturedKeys.insert(it->second)).second) {
                emit(attackerKeyStoredID, 1);
            }
        }
    }

    delete centre;
    return selectedNodes;
}

NodesSet NodeCompAttacker::centerDropCompromitation(int keysToBeCaptured, Coord * centre) {

    NodesSet selectedNodes;
    std::multimap<double, int> mapOfDistances;

    for (int i = 0; i < numNodes; i++) {
        cModule * node = FindModule<>::findNetwork(this)->getSubmodule("node", i);
        StationaryMobility * mobility = check_and_cast<StationaryMobility *>(node->getSubmodule("mobility"));
        Coord position = mobility->getCurrentPosition();

        mapOfDistances.insert(std::pair<double, int>(position.distance(*centre), i));
    }

    for (std::multimap<double, int>::iterator it = mapOfDistances.begin(); capturedKeys.size() < keysToBeCaptured && it != mapOfDistances.end(); it++) {
        cModule * node = FindModule<>::findNetwork(this)->getSubmodule("node", it->second);
        Data * data = check_and_cast<Data *>(node->getSubmodule("kms")->getSubmodule("data"));
        Keys keys = data->getKeys();

        if ((selectedNodes.insert(it->second)).second) {
            emit(comprNodeID, 1);
        }

        for (Keys::iterator it = keys.begin(); it != keys.end(); it++) {
            if ((capturedKeys.insert(it->second)).second) {
                emit(attackerKeyStoredID, 1);
            }
        }
    }

    delete centre;
    return selectedNodes;
}

NodeCompAttacker::~NodeCompAttacker() {

}
