#ifndef NODE_COMP_ATTACKER_H
#define NODE_COMP_ATTACKER_H

#include "BaseAttacker.h"

/**
 * @brief Implementation of Node compromise attacker
 *
 * Attacker compromises nodes based on selected compromise pattern. All keys
 * are exfiltrated from compromised nodes. Several nodes might be loaded with
 * malware.
 *
 * @author Radim Ostadal
 **/
class MIXIM_API NodeCompAttacker : public BaseAttacker {

    protected:
        /** @brief Pattern that will be used for initial compromise */
        int attackerCompPattern;

        /** @brief Stores all nodes compromised by the attacker */
        NodesSet compromisedNodes;

        /** @brief Different patterns used for initial compromise */
        enum attackerCompPatterns {
            RANDOM_PATTERN,
            OUTERMOST_PATTERN,
            DIRECT_CENTER_PATTERN,
            CENTER_DROP_PATTERN,
        };

    private:
        /** @brief Copy constructor is not allowed */
        NodeCompAttacker(const NodeCompAttacker&);

        /** @brief Assignment operator is not allowed */
        NodeCompAttacker& operator=(const NodeCompAttacker&);

    public:
        NodeCompAttacker();
        virtual ~NodeCompAttacker();

        /** @brief Initialization of variables, activity planned using timers */
        virtual void initialize(int stage);

        /**
         * @brief Attacker performs initial compromise
         *
         * Firstly, the number of keys that should be compromised is determined. Four
         * different compromise patterns are available based on attackerCompPattern
         * parameter. All keys are exfiltrated from compromised node. After the compromise,
         * several nodes are randomly loaded with a malware (particular number depends
         * on numInfectedNodes parameter).
         **/
        void performCompromitation();

        /** @brief Random nodes are selected and compromised up to the threshold of keysToBeCapture */
        NodesSet randomCompromitation(int keysToBeCapture);

        /** @brief The farthest nodes from the centre are compromised up to the threshold of keysToBeCapture */
        NodesSet outermostCompromitation(int keysToBeCapture, Coord * centre);

        /** @brief Attacker moves directly to a centre of the network from a random location on an edge of the network, picking up nodes in a close vicinity of his trajectory up to the threshold of keysToBeCapture */
        NodesSet directCenterCompromitation(int keysToBeCapture, Coord * centre);

        /** @brief The closest nodes to the centre of the network are compromised up to the threshold of keysToBeCapture */
        NodesSet centerDropCompromitation(int keysToBeCapture, Coord * centre);
};

#endif
