#include "Data.h"
#include "BaseSA.h"
#include "FindModule.h"
#include "utilfunctions.h"

Define_Module(Data);

Data::Data()
    : BaseApplLayer()
    , debug(0)
    , keyLength(0)
    , nonceLength(0)
    , headerLength(0)
    , processTableRep(0)
    , processTableInterval(0)
    , nonceConfPeriod(0)
    , attackerModule()
    , delayTimer()
    , confirmNoncesPeriodTimer()
    , predistributedKeys()
    , keys()
    , counters()
    , nonces()
    , activeNonces()
    , commonNeighbourDistAll()
{}

void Data::initialize(int stage) {
    BaseApplLayer::initialize(stage);

    if (stage == 0) {
        if (hasPar("debug")) debug = par("debug").boolValue();
        if (hasPar("keyLength")) keyLength = par("keyLength");
        if (hasPar("nonceLength")) nonceLength = par("nonceLength");
        if (hasPar("headerLength")) headerLength = par("headerLength");
        if (hasPar("processTableRep")) processTableRep = par("processTableRep");
        if (hasPar("processTableInterval")) processTableInterval = par("processTableInterval");
        if (hasPar("nonceConfPeriod")) nonceConfPeriod = par("nonceConfPeriod");

    } else if (stage == 1) {
        confirmNoncesPeriodTimer = new cMessage("Start processing table period", PROCESS_TABLE_PERIOD);
        scheduleAt(100 + uniform(0, 0.1, 9), confirmNoncesPeriodTimer);

        attackerModule = check_and_cast<BaseAttacker *>(FindModule<>::findNetwork(this)->getSubmodule("attacker"));
    }
}

void Data::handleLowerMsg(cMessage* msg) {
    if (msg->getKind() == SA_NONCE_PKT) {
        passIncNoncePkt(msg);

    } else if (msg->getKind() == NONCE_CONF_PKT) {
        processNonceIDsPkt(msg);

    } else {
        delete msg;
    }

}

void Data::handleSelfMsg(cMessage* msg) {
    if (msg->getKind() == PROCESS_TABLE_PERIOD) {
        confirmNoncesPeriodTimer = new cMessage("Start processing table period", PROCESS_TABLE_PERIOD);
        scheduleAt(simTime() + nonceConfPeriod, confirmNoncesPeriodTimer);

        processTable();

    } else if (msg->getKind() == PROCESS_TABLE) {
        processTable();
    }

    delete msg;
}

void Data::processTable() {

    Address neighbourAddress;

    // Decrement counters and clear structures
    if (decrementCountersAndClear()) {
        delayTimer = new cMessage("Process table", PROCESS_TABLE);
        scheduleAt(simTime() + processTableInterval + uniform(0, 0.1, 9), delayTimer);
    }

    // For the neighbours with higher IDs, do the movement from nonces to active nonces
    for (Counters::iterator it1 = counters.begin(); it1 != counters.end(); it1++) {
        neighbourAddress = it1->first;

        if (myApplAddr() < neighbourAddress) {

            // Move nonces from nonce structere to active nonce structure, active nonces have to be empty
            if (activeNonces[neighbourAddress].empty()) {

                // Iterate through all nonces for particular neighbour
                for (std::map<Hash, CountStruct>::iterator it2 = it1->second.begin(); it2 != it1->second.end(); it2++) {

                    Hash nonceHash = it2->first;

                    // Only nonces that are in nonces structure are considered (should be always true)
                    if (nonces[neighbourAddress].find(nonceHash) == nonces[neighbourAddress].end()) {continue;}

                    // Check 1 process table of waiting (to be sure that the neighbour also has a nonce)
                    if (it2->second.counter < (2 * processTableRep) + 2) {

                        // Move nonces from nonce structre to active nonces structure
                        Nonce nonce = nonces[neighbourAddress][nonceHash];
                        activeNonces[neighbourAddress][nonceHash] = nonce;
                        nonces[neighbourAddress].erase(nonceHash);

                        // Reset the counter for particular nonce (1rep for sending ID pkt, 1rep + 2 for receiving answers)
                        it2->second.counter = (2 * processTableRep) + 1;
                    }
                }
            }
        }
    }

    // Send the nonce ID packet (challenge or reply) for active nonces
    for (Nonces::iterator it1 = activeNonces.begin(); it1 != activeNonces.end(); it1++) {
        neighbourAddress = it1->first;

        // I am sending the challenge
        if (myApplAddr() < neighbourAddress && !activeNonces[neighbourAddress].empty()) {

            // Check whether to send the challenge or whether to wait for a response
            Hash testHash = (it1->second).begin()->first;
            if (counters[neighbourAddress][testHash].counter > processTableRep + 1) {

                // Prepare the nonce ID packet
                NonceIDsPkt *pkt = new NonceIDsPkt("Nonce confirmation challenge", NONCE_CONF_PKT);

                // Set source and destination address, and control info
                pkt->setDestAddr(neighbourAddress);
                pkt->setSrcAddr(myApplAddr());
                NetwControlInfo::setControlInfo(pkt, neighbourAddress);

                // Set the parameters
                pkt->setNonceHashesArraySize(activeNonces[neighbourAddress].size());

                // Insert all active hashes into the packet
                int i = 0;
                for (std::map<Hash, Nonce>::iterator it2 = activeNonces[neighbourAddress].begin(); it2 != activeNonces[neighbourAddress].end(); it2++) {
                    pkt->setNonceHashes(i, (it2->first));
                    i++;
                }

                // Send the packet for particular neighbour
                sendDelayedDown(pkt, uniform(0, processTableInterval, 9));

                emit(nonceConfChallengeSentID, 1);
            }
        // I am sending the reply
        } else if (myApplAddr() > neighbourAddress && !activeNonces[neighbourAddress].empty()){
            // Prepare the nonce ID packet
            NonceIDsPkt *pkt = new NonceIDsPkt("Nonce confirmation reply", NONCE_CONF_PKT);

            // Set source and destination address, and control info
            pkt->setDestAddr(neighbourAddress);
            pkt->setSrcAddr(myApplAddr());
            NetwControlInfo::setControlInfo(pkt, neighbourAddress);

            // Set the parameters
            pkt->setNonceHashesArraySize(activeNonces[neighbourAddress].size());

            // Insert all active hashes into the packet
            int i = 0;
            for (std::map<Hash, Nonce>::iterator it2 = activeNonces[neighbourAddress].begin(); it2 != activeNonces[neighbourAddress].end(); it2++) {
                pkt->setNonceHashes(i, (it2->first));
                i++;
            }

            // Send the packet for particular neighbour
            sendDelayedDown(pkt, uniform(0, processTableInterval, 9));

            emit(nonceConfResponseSentID, 1);
        }
    }
}

bool Data::decrementCountersAndClear() {
    bool somethingToProcess = 0;
    std::map<Address, HashSet> toBeDeleted;

    for (Counters::iterator it1 = counters.begin(); it1 != counters.end(); it1++) {
        for (std::map<Hash, CountStruct>::iterator it2 = it1->second.begin(); it2 != it1->second.end(); it2++) {

            if (it2->second.counter == 0) {
                // Set the record to be deleted at the end of function
                toBeDeleted[it1->first].insert(it2->first);
            } else {
                // Decrement counter
                it2->second.counter--;
                // There is still something to process, it is necessary to plan next processing of table
                somethingToProcess = 1;
            }
        }
    }

    // Clear all data for expired counters
    for (std::map<Address, HashSet>::iterator it1 = toBeDeleted.begin(); it1 != toBeDeleted.end(); it1++) {
        for (HashSet::iterator it2 = it1->second.begin(); it2 != it1->second.end(); it2++) {
            counters[it1->first].erase(*it2);
            nonces[it1->first].erase(*it2);
            activeNonces[it1->first].erase(*it2);
        }
    }

    return somethingToProcess;
}

void Data::processNonceIDsPkt(cMessage* msg) {
    NonceIDsPkt *pkt  = check_and_cast<NonceIDsPkt*>(msg);
    Address neighbourAddress = pkt->getSrcAddr();

    if (myApplAddr() > neighbourAddress) {
        // The packet is challenge from my neighbour (I have higher ID)
        // Do the key update and move the nonces from nonces structure to active nonces

        emit(nonceConfChallengeReceivedID, 1);

        for (int i = 0; i < pkt->getNonceHashesArraySize(); i++) {
            Hash nonceHash = pkt->getNonceHashes(i);

            // If I have the same hash for my neighbour, update key with this hash and move it to active nonces
            if (nonces[neighbourAddress].find(nonceHash) != nonces[neighbourAddress].end()) {

                // Update key
                updateKey(neighbourAddress, nonces[neighbourAddress][nonceHash]);

                // Move nonces from nonce structre to active nonces structure
                Nonce nonce = nonces[neighbourAddress][nonceHash];
                activeNonces[neighbourAddress][nonceHash] = nonce;
                nonces[neighbourAddress].erase(nonceHash);

                // Reset the counter for particular nonce
                // Lifetime of active nonce for sending replies is 1 repetitions
                counters[neighbourAddress][nonceHash].counter = processTableRep;
            }
        }
    } else {
        // The packet is respond to my challenge, do the key update
        emit(nonceConfResponseReceivedID, 1);

        for (int i = 0; i < pkt->getNonceHashesArraySize(); i++) {
            Hash nonceHash = pkt->getNonceHashes(i);

            // I have to have the nonce in active nonces structure
            if (activeNonces[neighbourAddress].find(nonceHash) != activeNonces[neighbourAddress].end()) {

                // If the key was not already updated with this nonce, do update
                if (!counters[neighbourAddress][nonceHash].keyUpdated) {

                    // Update key
                    updateKey(neighbourAddress, activeNonces[neighbourAddress][nonceHash]);

                    // Set the paramater that the key was already updated
                    counters[neighbourAddress][nonceHash].keyUpdated = true;
                }
            }
        }
    }

    delete msg;
}

void Data::insertNonce(Address addr, Nonce nonce) {

    // Check that the nonce is for my neighbour - I do have a key for this node
    if (!isMyNeighbour(addr)) {return;}

    // Compute hash for nonce
    unsigned int nonceHash;
    nonceHash = SHA256int(nonce, nonceLength);

    // Insert nonce to the nonces table
    nonces[addr][nonceHash] = nonce;

    // Insert nonce to the counters table
    if (myApplAddr() < addr) {
        // I will be the one who process / challenge this nonce
        // 1 - initial waiting - to be sure that my neighbour also received nonce
        // 1rep - I will be sending the challenge
        // 1rep - I will be waiting for the response
        // 2 - particular nodes are not synchronised and interval for sending is [0, processTableInterval]
        counters[addr][nonceHash].counter = (2 * processTableRep) + 3;
    } else {
        // My neighbour will challenge the nonce
        // 2 times the previous number in the worst case, than the noce might be deleted
        counters[addr][nonceHash].counter = (4 * processTableRep) + 6;
    }

    counters[addr][nonceHash].keyUpdated = false;
}

void Data::updateKey(Address addr, Nonce nonce) {
    unsigned char buffer[64];

    memcpy(buffer, keys[addr], 32);
    memcpy(buffer + 32, nonce, 32);

    Key newKey = SHA256(buffer, 64);
    Key oldKey = keys[addr];

    // Save the new key for the neighbour
    keys[addr] = newKey;

    // Check the new key
    attackerModule->attackerKeyUpdateCheck(oldKey, nonce, newKey);
}

void Data::passOutNoncePkt(cMessage* msg, simtime_t delay) {
    Enter_Method_Silent();
    take(msg);

    sendDelayedDown(msg, delay);
}

void Data::passIncNoncePkt(cMessage* msg) {
    BaseSA* sa = check_and_cast<BaseSA *>(getParentModule()->getSubmodule("sa"));
    sa->processNoncePkt(msg);

}

void Data::setPredistributedKey(Address addr, Key key) {
    predistributedKeys[addr] = key;
}

void Data::setKey(Address addr, Key key) {
    keys[addr] = key;
}

bool Data::isMyNeighbour(Address addr) {
    if (keys.find(addr) != keys.end()) {
        return true;
    } else {
        return false;
    }
}

void Data::setCommonNeighbourDist(Address NPaddr, Address addr, double distance_nc, double distance_np) {
    commonNeighbourDistAll[NPaddr][addr].first = distance_nc;
    commonNeighbourDistAll[NPaddr][addr].second = distance_np;
}

Data::~Data() {

}
