#ifndef DATA_H
#define DATA_H

#include "BaseApplLayer.h"
#include "kmsdefs.h"
#include "NonceIDsPkt_m.h"
#include "BaseAttacker.h"

/**
 * @brief Data class is responsible to store all the data on the node; it is
 * also responsible for nonce confirmation and link keys update
 *
 * All the nonces that are received by the node are stored in Data class. The
 * repetitive process processTable handles the nonce lifecycle. The lifecycle
 * consists of wait period (to be sure, that neighbouring node also received
 * the nonce), nonce confirmation challenges and responses, and the key update.
 * Other setting and getting functions for keys, predistributed keys, etc. are
 * defined.
 *
 * @author Radim Ostadal
 **/
class Data : public BaseApplLayer {
    protected:
        /** @brief Debug switch */
        bool debug;

        /** @brief Length of the secret key (in bytes) */
        int keyLength;

        /** @brief Length of the nonce (in bytes) */
        int nonceLength;

        /** @brief Length of the network packet header (in bits) */
        int headerLength;

        /** @brief Every nonce confirmation challenge and response are sent processTableRep times */
        int processTableRep;

        /** @brief Defines how often the table is processed (planning the nonce confirmation challenges and responses) */
        int processTableInterval;

        /** @brief The interval between two nonce confirmation periods */
        int nonceConfPeriod;

        /** @brief Pointer to attacker module */
        BaseAttacker * attackerModule;

        /** @brief Timer for planing the process table event */
        cMessage * delayTimer;

        /** @brief Timer for planing the nonce confirmation period */
        cMessage * confirmNoncesPeriodTimer;

        /** @brief Array with predistributed keys (provided by predistribution centre) */
        Keys predistributedKeys;

        /** @brief Array for current keys for all node's neighbours */
        Keys keys;

        /** @brief The primary structure to store control information about every nonce for my neighbours */
        Counters counters;

        /** @brief The structure for nonce storing after they are received */
        Nonces nonces;

        /** @brief The structure for nonce storing, when they are being used for mutual confirmation and key update */
        Nonces activeNonces;

        /** @brief The structure stores all mutual neighbours for all nodes's neighbours and respective distances for hybrid designed protocols */
        CommonNeighbourDistAll commonNeighbourDistAll;

        // All signals for collection of statistics
        simsignal_t nonceConfChallengeSentID = registerSignal("nonceConfChallengeSent");
        simsignal_t nonceConfResponseSentID = registerSignal("nonceConfResponseSent");
        simsignal_t nonceConfChallengeReceivedID = registerSignal("nonceConfChallengeReceived");
        simsignal_t nonceConfResponseReceivedID = registerSignal("nonceConfResponseReceived");

        /** @breif Message types used for timers */
        enum selfMessages {
            PROCESS_TABLE,
            PROCESS_TABLE_PERIOD,
        };

    private:
        /** @brief Copy constructor is not allowed */
        Data (const Data&);

        /** @brief Assignment operator is not allowed */
        Data& operator=(const Data&);

    public:
        Data();
        virtual ~Data();

        /** @brief Initialization of variables, activity planned using timers */
        void initialize(int stage);

        /** @brief Function handling secrecy amplification and nonce confirmation messages */
        void handleLowerMsg(cMessage* msg);

        /** @brief Handle the self messages - processTable planning */
        void handleSelfMsg(cMessage *msg);

        /**
         * @brief Function responsible for handling the nonce confirmation period and nonce lifecycle
         *
         * Firstly, all counters are decremented and expired nonces are deleted (decrementCountersAndClear
         * function). The node with lower id is responsible for sending the nonce confirmation request.
         * The confirmation request is sent for multiple nonces in one message (using the nonce IDs). If the
         * node has bigger id, the nonce is waiting for the confirmation request from a neighbour. After the
         * reception, confirmation responses are being send. If there is no confirmation request,
         * the nonce expires in some time and is deleted. After the nonce is mutually confirmed, it is
         * used to update the link key.
         **/
        void processTable();

        /** @brief Function decrement counter (life time) for all nonces and delete the expired ones */
        bool decrementCountersAndClear();

        /**
         * @brief Update the link key for neighbour (identified by addr parameter) with nonce; the
         * SHA-256 is used on original key || nonce
         */
        void updateKey(Address addr, Nonce nonce);

        /** @brief Setting function for predistributed keys */
        void setPredistributedKey(Address addr, Key key);

        /** @brief Setting function for current link keys */
        void setKey(Address addr, Key key);

        /**
         * @brief Function setting the common neighbours with all my neighbours and respective
         * distances used for intermediate node selection in hybrid designed protocols.
         *
         * @param NPaddr Address of my neighbouring node NP
         * @param addr Address of the node that is common neighbour of NP and myself
         * @param distance_nc The absolute distance of node addr from myself
         * @param distance_np The absolute distance of node addr from NP
         **/
        void setCommonNeighbourDist(Address NPaddr, Address addr, double distance_nc, double distance_np);

        /** @brief Return the key length in bytes */
        int getKeyLength() {return keyLength;};

        /** @brief Return the predistributed key for the node identified by addr parameter */
        Key getPredistributedKey(Address addr) {return predistributedKeys[addr];};

        /** @brief Return the current key for the node identified by addr parameter */
        Key getKey(Address addr) {return keys[addr];};

        /** @brief Return all predistributed keys for all nodes */
        Keys getPredistributedKeys() {return predistributedKeys;};

        /** @brief Return all currently used keys for all node's neighbours */
        Keys getKeys() {return keys;};

        /** @brief Return all nonces stored in the nonces structure for all node's neighbours */
        Nonces getNonces() {return nonces;};

        /**
         * @brief Return all our common neighbours with neighbouring node identified by addr parameter
         * including the absolute distances
         */
        CommonNeighbourDist getCommonNeighbourDist(Address addr) {return commonNeighbourDistAll[addr];};

        /** @brief Return all common neighbours with all of my neighbours including the absolute distances. */
        CommonNeighbourDistAll getCommonNeighbourDistAll() {return commonNeighbourDistAll;};

        /** @brief Function used by SA module to pass the nonces to a network layer */
        void passOutNoncePkt(cMessage* msg, simtime_t delay);

        /** @brief Function used to pass the nonce to SA module from a network layer */
        void passIncNoncePkt(cMessage* msg);

        /** @brief Function processing the nonce confirmation message from a neighbour */
        void processNonceIDsPkt(cMessage* msg);

        /** @brief Function used by SA module to insert the new nonce for a particular neighbour identified by addr parameter */
        void insertNonce(Address addr, Nonce nonce);

        /** @brief Return true if the node identified by addr parameter is my neighbour, false otherwise */
        bool isMyNeighbour(Address addr);

        /**
         * @brief Message types processed by the Data module
         *
         * SA_NONCE_PKT is used for nonce packets, NONCE_CONF_PKT is used for confirmations messages.
         **/
        enum pktMessageKinds {
            SA_NONCE_PKT,
            NONCE_CONF_PKT,
        };
};

#endif
