#ifndef BASE_SA_H
#define BASE_SA_H

#include "BaseApplLayer.h"
#include "NetwControlInfo.h"
#include "Data.h"
#include "utilfunctions.h"
#include "NoncePkt_m.h"

/**
 * @brief Base class for implementation of secrecy amplification protocols
 *
 * The common functions for all secrecy amplification protocols are implemented.
 * This includes processing and resending the nonce messages, starting the
 * nonce distribution period, etc.
 *
 * @author Radim Ostadal
 **/
class MIXIM_API BaseSA : public BaseApplLayer {

    protected:
        /** @brief Debug switch */
        bool debug;

        /** @brief Length of the network packet header (in bits) */
        int headerLength;

        /** @brief Length of nonce distribution phase */
        int noncePeriod;

        /** @brief The interval between two nonce distribution periods */
        int amplifRepetitions;

        /** @brief Number of rounds of nonce distribution */
        int amplifPeriod;

        /** @brief Specification of protocol used for amplification */
        int protocolType;

        /** @brief Pointer to data modul */
        Data * dataModule;

        /** @brief Timer used to execute the particular protocol */
        cMessage * delayTimer;

        // All signals for collection of statistics
        simsignal_t amplifNonceGeneratedID = registerSignal("amplifNonceGenerated");
        simsignal_t amplifNonceSentID = registerSignal("amplifNonceSent");
        simsignal_t amplifNonceSavedID = registerSignal("amplifNonceSaved");
        simsignal_t amplifNonceReSentInter1ID = registerSignal("amplifNonceReSentInter1");
        simsignal_t amplifNonceDiscardInter1ID = registerSignal("amplifNonceDiscardInter1");
        simsignal_t amplifNonceReSentInter2ID = registerSignal("amplifNonceReSentInter2");
        simsignal_t amplifNonceDiscardInter2ID = registerSignal("amplifNonceDiscardInter2");
        simsignal_t amplifNonceReceivedTotalID = registerSignal("amplifNonceReceivedTotal");
        simsignal_t amplifNonceReceivedMultihoplID = registerSignal("amplifNonceReceivedMultihopl");
        simsignal_t amplifNonceReceivedDirectID = registerSignal("amplifNonceReceivedDirect");

        /** @breif Message types used for timers */
        enum selfMessages {
            EXECUTE_PROTOCOL,
            INSERT_NONCE,
        };

    private:
        /** @brief Copy constructor is not allowed */
        BaseSA (const BaseSA&);

        /** @brief Assignment operator is not allowed */
        BaseSA& operator=(const BaseSA&);

    public:
        BaseSA();
        virtual ~BaseSA();

        /** @brief Initialization of variables, plan the start of protocol execution */
        virtual void initialize(int stage);

        /** @brief Handle the self messages - starts the secrecy amplification activity */
        void handleSelfMsg(cMessage* msg);

        /** @brief Virtual function that has to be implemented in actual class performing secrecy amplification */
        virtual void executeProtocol();

        /**
         * @brief Processing nonce packet: insert the nonce into the Data structure if the
         * packet is for me, resend otherwise
         **/
        void processNoncePkt(cMessage* msg);
};

#endif
