#include "HybridDesign.h"

Define_Module(HybridDesign);

HybridDesign::HybridDesign()
    : BaseSA()
    , nodeTransmissionRange(0)
{}

void HybridDesign::initialize(int stage) {
    BaseSA::initialize(stage);

    if (stage == 0) {
        if (hasPar("protocolType")) {
            if (!strcmp(par("protocolType"), "HDFINAL")) protocolType = HDFINAL_NONCE;
            if (!strcmp(par("protocolType"), "HDBEST")) protocolType = HDBEST_NONCE;
        }
        if (hasPar("nodeTransmissionRange")) nodeTransmissionRange = par("nodeTransmissionRange");

    } else if (stage == 1) {

    }
}

void HybridDesign::executeProtocol() {
    if (protocolType == HDFINAL_NONCE) {
        performHDFinal();
    } else if (protocolType == HDBEST_NONCE) {
        performHDBest();
    }
}

void HybridDesign::performHDFinal() {
    // HD Final protocol intermediate nodes: 0.69_0.98 and 0.01_0.39

    Keys keys = dataModule->getKeys();
    Address dstNode;

    // Iterate through all my neighbours (take destination node)
    for (Keys::iterator it1 = keys.begin(); it1 != keys.end(); it1++) {
        dstNode = it1->first;

        basicPush(dstNode, findInterNode(dstNode, 0.69, 0.98));
        basicPush(dstNode, findInterNode(dstNode, 0.01, 0.39));
    }
}

void HybridDesign::performHDBest() {
    // HD Final protocol intermediate nodes: 0.32_0.85, 0.56_0.70, 0.89_0.01, 0.69_0.98 and 0.01_0.39

    Keys keys = dataModule->getKeys();
    Address dstNode;

    // Iterate through all my neighbours (take destination node)
    for (Keys::iterator it1 = keys.begin(); it1 != keys.end(); it1++) {
        dstNode = it1->first;

        basicPush(dstNode, findInterNode(dstNode, 0.69, 0.98));
        basicPush(dstNode, findInterNode(dstNode, 0.01, 0.39));
        basicPush(dstNode, findInterNode(dstNode, 0.32, 0.85));
        basicPush(dstNode, findInterNode(dstNode, 0.56, 0.70));
        basicPush(dstNode, findInterNode(dstNode, 0.89, 0.01));
    }
}

void HybridDesign::basicPush(Address dstNode, Address interNode2) {

    NoncePkt *pkt = new NoncePkt("Basic push nonce", Data::SA_NONCE_PKT);
    NoncePkt *selfPkt = new NoncePkt("Self push nonce", INSERT_NONCE);

    // Generate random nonce
    int keyLength = dataModule->getKeyLength();
    Nonce nonce = new unsigned char [keyLength];
    rndArray(nonce, keyLength);

    emit(amplifNonceGeneratedID, 1);

    // Set nonce to packet
    pkt->setNonceArraySize(keyLength);
    selfPkt->setNonceArraySize(keyLength);
    for (int i = 0; i < keyLength; i++) {
        pkt->setNonce(i, nonce[i]);
        selfPkt->setNonce(i, nonce[i]);
    }

    // Set the source node, destination node, neighbour node and intermediate node
    pkt->setDstNode(dstNode);
    pkt->setSrcNode(myApplAddr());
    pkt->setNeighbourNode(myApplAddr());
    pkt->setInterNode2(interNode2);

    // Generate the delay for the packets
    double pktDelay = uniform(0, noncePeriod);

    // Set the neigbouring node for self msg (my neighbour for which the nonce was generated)
    selfPkt->setNeighbourNode(dstNode);

    // Schedule the self packet
    scheduleAt(simTime() + pktDelay, selfPkt);

    emit(amplifNonceSavedID, 1);

    // Send the packet
    NetwControlInfo::setControlInfo(pkt, interNode2);

    dataModule->passOutNoncePkt(pkt, pktDelay);
    emit(amplifNonceSentID, 1);
}

Address HybridDesign::findInterNode(Address neighAddr, double centralRelativeIdentif, double neighRelativeIdentif) {

    // Compute the relative distances from protocol definition
    double centralRelDist = centralRelativeIdentif * nodeTransmissionRange;
    double neighRelDist = neighRelativeIdentif * nodeTransmissionRange;

    double currentDistance = 0;
    double minimalDistance = 2 * pow(nodeTransmissionRange, 2);

    // In case there is no better neighbour, use the direct link
    Address returnNode = neighAddr;

    CommonNeighbourDist commonNeighbourDist = dataModule->getCommonNeighbourDist(neighAddr);

    for (CommonNeighbourDist::iterator it1 = commonNeighbourDist.begin(); it1 != commonNeighbourDist.end(); it1++) {

        currentDistance = 0;
        currentDistance += pow(it1->second.first - centralRelDist, 2);
        currentDistance += pow(it1->second.second - neighRelDist, 2);

        if (currentDistance <= minimalDistance) {
            minimalDistance = currentDistance;
            returnNode = it1->first;
        }
    }

    return returnNode;
}

HybridDesign::~HybridDesign() {

}
