#ifndef HYBRID_DESIGN_H
#define HYBRID_DESIGN_H

#include "BaseSA.h"

/**
 * @brief Implementation of Hybrid designed protocols.
 *
 * The class provides with implementation of two different hybrid protocols: HD Final
 * and HD Best.
 *
 * @author Radim Ostadal
 **/
class MIXIM_API HybridDesign : public BaseSA {
    protected:
        /** @brief Definition of legitimate node's transmission range in metres */
        double nodeTransmissionRange;

        /** @breif Protocol types used within a class */
        enum protocolTypes {
            HDFINAL_NONCE,
            HDBEST_NONCE,
        };

    private:
        /** @brief Copy constructor is not allowed */
        HybridDesign(const HybridDesign&);

        /** @brief Assignment operator is not allowed */
        HybridDesign& operator=(const HybridDesign&);

    public:
        HybridDesign();
        virtual ~HybridDesign();

        /** @brief Initialization of variables, activity planned using timers */
        virtual void initialize(int);

        /** @brief The particular protocol is executed based on protocolType parameter */
        void executeProtocol();

        /**
         * @brief Implementation of HD Final protocol
         *
         * The protocol is a variant of Push protocol, where only a defined subset of
         * intermediate nodes is used to forward a nonce. Node successively
         * selects a destination node B from its neighbours. For every node B, two
         * intermediate nodes C1 and C2 are selected from common neighbours of myself and B.
         * A selection is done based on relative distances.
         * Two generated nonces are sent to node B, first via node C1 and the second one via
         * node C2.
         **/
        void performHDFinal();

        /**
         * @brief Implementation of HD Best protocol
         *
         * The best hybrid designed protocol is variant of HD Final protocol. The only
         * difference is that five intermediate nodes C1..C5 are selected instead of two.
         **/
        void performHDBest();

        /**
         *  @brief The function takes the destination and intermediate node; it generates
         *  a new nonce and sends it to the destination node through the intermediate node
         **/
        void basicPush(Address dstNode, Address interNode2);

        /**
         * @brief Returns the address of neighbouring node with a suitable position
         *
         * @param neighAddr Address of neighbouring node; return node will be our common neighbour
         * @param centralRelativeIdentif The relative distance from myself
         * @param neighRelativeIdentif The relative distance from neighAddr node
         **/
        Address findInterNode(Address neighAddr, double centralRelativeIdentif, double neighRelativeIdentif);
};

#endif
