#ifndef NODE_ORIENTED_H
#define NODE_ORIENTED_H

#include "BaseSA.h"

/**
 * @brief Implementation of Node oriented protocols
 *
 * The class provides with implementation of five different node oriented protocols:
 * Pull, Push, Multi-hop Pull, Multi-hop Push, and NO Best.
 *
 * @author Radim Ostadal
 **/
class MIXIM_API NodeOriented : public BaseSA {
    protected:
        /** @breif Protocol types used within a class */
        enum ProtocolTypes {
            PULL_NONCE,
            PUSH_NONCE,
            MULTI_PULL_NONCE,
            MULTI_PUSH_NONCE,
            NO_BEST_ORIG_NONCE,
            NO_BEST_NEW_NONCE,
        };

    private:
        /** @brief Copy constructor is not allowed */
        NodeOriented(const NodeOriented&);

        /** @brief Assignment operator is not allowed */
        NodeOriented& operator=(const NodeOriented&);

    public:
        NodeOriented();
        virtual ~NodeOriented();

        /** @brief Initialization of variables, activity planned using timers */
        virtual void initialize(int);

        /** @brief The particular protocol is executed based on protocolType parameter */
        void executeProtocol();

        /**
         * @brief Implementation of Pull protocol
         *
         * The node selects two destination nodes A and B from its neighbours repetitivly.
         * A unique nonce is generated for all pairs of nodes A and B. The nonces are sent
         * to both nodes.
         **/
        void performPull();

        /**
         * @brief Implementation of Push protocol
         *
         * Sensor node selects a destination node B and an intermediate node C from its
         * neighbours. A unique nonce is generated for all combinations of destination and
         * intermediate nodes. The node sends generated nonce via node C to node B.
         **/
        void performPush();

        /**
         * @brief Implementation of Multi-hop Pull protocol
         *
         * Sensor node selects two destination nodes A and B, and one intermediate node D
         * from its neighbours. A unique nonce is generated for all combinations of nodes
         * A, B and D. The nonce is sent directly to node A and via node D to node B.
         **/
        void performMultiPull();

        /**
         * @brief Implementation of Multi-hop Push protocol
         *
         * Sensor node selects a destination node B and two intermediate nodes C and D from
         * its neighbours. A unique nonce is generated for all combinations of nodes B, C
         * and D. The nonce is sent to node D via node C. Finally, node D forward the nonce
         * to the destination node B.
         **/
        void performMultiPush();
};

#endif
